<?php
/**
 * Anwiki is a multilingual content management system <http://www.anwiki.com>
 * Copyright (C) 2007-2008 Antoine Walter <http://www.anw.fr>
 * 
 * Anwiki is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 * 
 * Anwiki is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Anwiki.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * ContentField definition, and implementation of various common ContentFields.
 * @package Anwiki
 * @version $Id: class_contentfield.php,v 1.23 2008-09-11 22:42:02 antoine-cvs Exp $
 * @copyright 2007-2008 Antoine Walter
 * @license http://www.gnu.org/copyleft/gpl.html GNU Public License 3
 */

/* In fact, it's AnwContentField*_container which should inherit from AnwStructuredContentFieldsContainer.
 * This is not possible untill multiple inherit is supported by PHP.
 */
abstract class AnwStructuredContentField extends AnwStructuredContentFieldsContainer
{
	private $sName;
	private $oMultiplicity;
	private $bCollapsed;
	private $oContainer; //parent (contentfield or contentclass)
	private $amCachedFieldTranslations = array();
	private $amDefaultValues;
	private $oRelatedComponent; //for looking up translations of contentfields declared from components (instead of declared by contentclass)
	private $bMandatory = false;
	
	private static $bDisplayOnlyMandatory = false;
	private static $bDoCollapsing = true;
	
	const INPUT_PREFIX = "anwpost";
	
	function __construct($sName)
	{
		$this->sName = $sName;
		
		if ($this instanceof AnwStructuredContentField_renderedAsMultiple)
		{
			$this->oMultiplicity = new AnwContentMultiplicity_multiple();
		}
		else
		{
			$this->oMultiplicity = new AnwContentMultiplicity_single();
		}
		
		if ($this instanceof AnwStructuredContentField_composed)
		{
			$this->setCollapsed(true);
		}
	}
	
	//overridable
	function init()
	{
		
	}
	
	function setRelatedComponent($oRelatedComponent)
	{
		$this->oRelatedComponent = $oRelatedComponent;
	}
	
	function getRelatedComponent()
	{
		if ($this->oRelatedComponent)
		{
			return $this->oRelatedComponent;
		}
		
		if ($this->getContainer() instanceof AnwStructuredContentField_composed)
		{
			return $this->getContainer()->getRelatedComponent();
		}
		
		return null;
	}
	
	function setMultiplicity($oMultiplicity)
	{
		$this->oMultiplicity = $oMultiplicity;
		
		if ($this instanceof AnwStructuredContentField_renderedAsMultiple 
			&& !($oMultiplicity instanceof AnwContentMultiplicity_multiple))
		{
			throw new AnwUnexpectedException("content field _renderedAsMultiple must have multiple multiplicity!");
		}
		
		if ($oMultiplicity instanceof AnwContentMultiplicity_multiple 
			&& !$this->amDefaultValues)
		{
			$this->setDefaultValues(array());
		}
	}
	
	function setMandatory($bMandatory)
	{
		$this->bMandatory = $bMandatory;
	}
	
	function isMandatory()
	{
		return $this->bMandatory;
	}
	
	function setCollapsed($bCollapsed)
	{
		$this->bCollapsed = $bCollapsed;
	}
	
	function getName()
	{
		return $this->sName;
	}
	
	function getMultiplicity()
	{
		return $this->oMultiplicity;
	}
	
	function getContainer()
	{
		return $this->oContainer;
	}
	
	function setDefaultValues($amValues)
	{
		$this->amDefaultValues = $amValues;
	}
	
	function getDefaultValues()
	{
		if ($this->isMultiple())
		{
			if ($this->amDefaultValues)
			{
				return $this->amDefaultValues;
			}
		}
		else
		{
			//atomic field
			if ($this->hasDefaultValue())
			{
				return array($this->getDefaultValue());
			}
		}
		throw new AnwUnexpectedException("no default values");
	}
	
	function hasDefaultValues()
	{
		if ($this->isMultiple())
		{
			return ($this->amDefaultValues ? true : false);
		}
		else
		{
			return $this->hasDefaultValue();
		}
	}
	
	/**
	 * Warning! This function may call overloaded functions by contentfields with tests consomming high cpu time
	 * (such as connecting to a database for checking that valid user/pwd have been edited).
	 * This function should be only called when it's really needed, and should never be called more than one time.
	 */
	final function testContentFieldValues($asFieldValues)
	{
		//test multiplicity
		$this->testContentFieldMultiplicity($asFieldValues);
	
		if ($this->isMandatory() && !$this->isMultiple())
		{
			$sValue = reset($asFieldValues);
			if ($sValue == $this->getDefaultValue())
			{
				$sError = AnwComponent::g_editcontent("err_contentfield_mandatory");
				throw new AnwInvalidContentFieldValueException($sError);
			}
		}
		
		//test values
		foreach ($asFieldValues as $sFieldValue)
		{
			/*//check PHP permission
			if (AnwUtils::contentHasPhpCode($sFieldValue) && !AnwCurrentSession::getUser()->isPhpEditionAllowed())
			{
				//just to avoid evaluating php syntax
				$sError = AnwComponent::g_editcontent("err_contentfield_acl_php");
				throw new AnwInvalidContentFieldValueException($sError);
			}
			*/
			
			//here we don't return a simple 'AnwInvalidContentFieldValueException' to prevent unauthorized users to access PHP source
			//this will display a big ACL error page instead of edit form...			
			if (AnwUtils::contentHasPhpCode($sFieldValue))
			{
				AnwCurrentSession::getUser()->checkPhpEditionAllowed();
			}
			
			//check JS permission
			if (AnwUtils::contentHasJsCode($sFieldValue) && !AnwCurrentSession::getUser()->isJsEditionAllowed())
			{
				$sError = AnwComponent::g_editcontent("err_contentfield_acl_js");
				throw new AnwInvalidContentFieldValueException($sError);
			}
			
			
			$this->testContentFieldValue($sFieldValue);
		}
	}
	
	//overridable
	protected function testContentFieldMultiplicity($asFieldValues)
	{
		$this->getMultiplicity()->testContentFieldMultiplicity($this, $asFieldValues);
	}
	
	abstract function testContentFieldValue($mContentFieldValue);
	
	final function getInputName($sSuffix)
	{
		$sInputName = self::INPUT_PREFIX.$sSuffix.'-'.$this->getName();
		return $sInputName;
	}	
	
	protected final function getInputId($sInputName, $nInputNumber)
	{
		return $sInputName.'-'.$nInputNumber;
	}
	
	final function getValueFromPost($sSuffix)
	{
		if (!($this->getMultiplicity() instanceof AnwContentMultiplicity_single)) throw new AnwUnexpectedException("getValueFromPost called on non single multiplicity !");
		$asFieldValues = $this->getValuesFromPost($sSuffix);
		return $asFieldValues[0];
	}
	
	function setContainer($oContentFieldContainer)
	{
		$this->oContainer = $oContentFieldContainer;
	}
	
	final function getContentClass()
	{
		if ($this->getContainer() instanceof AnwStructuredContentClass)
		{
			return $this->getContainer();
		}
		else if ($this->getContainer() instanceof AnwStructuredContentField_composed)
		{
			return $this->getContainer()->getContentClass();
		}
		else
		{
			//should never go here
			throw new AnwUnexpectedException("getContentClass : not a contentclass or a contentfield composed");
		}
	}
	
	abstract function getComponent();
	
	
	static function setDisplayOnlyMandatory($bDisplayOnlyMandatory)
	{
		self::$bDisplayOnlyMandatory = $bDisplayOnlyMandatory;
	}
	static function getDisplayOnlyMandatory()
	{
		return self::$bDisplayOnlyMandatory;
	}
	function hasMandatoryChild()
	{
		if ($this instanceof AnwStructuredContentField_composed)
		{
			$aoSubContentFields = $this->getContentFields();
			foreach ($aoSubContentFields as $oSubContentField)
			{
				if ($oSubContentField->isMandatory() || $oSubContentField->hasMandatoryChild())
				{
					return true;
				}
			}
		}
		return false;
	}	
	function isMandatoryChild()
	{
		return ($this->getContainer() instanceof AnwStructuredContentField_composed 
				&& ($this->getContainer()->isMandatory() || $this->getContainer()->isMandatoryChild()));
	}
	function isDisplayed()
	{
		if (self::getDisplayOnlyMandatory() && !$this->isMandatory() && !$this->hasMandatoryChild() &&!$this->isMandatoryChild())
		{
			return false;
		}
		return true;
	}
	
	
	static function setDoCollapsing($bCollapsing)
	{
		self::$bDoCollapsing = $bCollapsing;
	}
	
	static function getDoCollapsing()
	{
		return self::$bDoCollapsing;
	}
	
	function isCollapsed()
	{
		//TODO: for the moment, a contentfield who has collapsed childs can't be collapsed.
		//otherwise, we would encounter serious edition problems as same edit inputs would appear twice in html source.
		if (self::getDoCollapsing() && $this->bCollapsed && !$this->hasCollapsedChild())
		{
			return true;
		}
		return false;
	}
	
	function isCollapsedChild()
	{
		return ($this->getContainer() instanceof AnwStructuredContentField_composed 
				&& $this->getContainer()->isCollapsed());
	}
	
	function hasCollapsedChild()
	{
		if ($this instanceof AnwStructuredContentField_composed)
		{
			$aoSubContentFields = $this->getContentFields();
			foreach ($aoSubContentFields as $oSubContentField)
			{
				if ($oSubContentField->isCollapsed() || $oSubContentField->hasCollapsedChild())
				{
					return true;
				}
			}
		}
		return false;
	}
	
	abstract protected function renderEditInput($sInputValue, $sSuffix, $nInputNumber);
	abstract protected function renderCollapsedInput($sInputValue, $sSuffix);
	
	final function getRenderedEditInput($sInputValue, $sSuffix, $nInputNumber)
	{
		if ($this->isCollapsed() || $this->isCollapsedChild())
		{
			if ($this instanceof AnwStructuredContentField_composed)
			{
				$sInputValueUncollapsed = array(
					AnwStructuredContent::IDX_SUBCONTENT=>$sInputValue[AnwStructuredContent::IDX_SUBCONTENT],
					AnwStructuredContent::IDX_RENDERED=>$sInputValue['UNCOLLAPSED']);
				$sUncollapsedHtml = $this->renderEditInput($sInputValueUncollapsed, $sSuffix, $nInputNumber);
				
				$sCollapsedHtml = $this->renderCollapsedInput($sInputValue, $sSuffix);
				
				return $this->getComponent()->tpl()->renderCollapsedInput($sUncollapsedHtml, $sCollapsedHtml, $this);
			}
			else
			{
				if ($this instanceof AnwStructuredContentField_renderedAsMultiple)
				{
					$sHtml = "";
					foreach ($sInputValue as $sOneValue)
					{
						$sHtml .= $this->renderCollapsedInput($sOneValue, $sSuffix);
					}
					return $sHtml;
				}
				return $this->renderCollapsedInput($sInputValue, $sSuffix);
			}
		}
		
		return $this->renderEditInput($sInputValue, $sSuffix, $nInputNumber);
	}
	
	final function renderEditInputs($amValues, $sSuffix, $sFieldError=false)
	{
		$sInputHtml = $this->getMultiplicity()->doRenderEditInputs($this, $amValues, $sSuffix);
		
		$sFieldTip = "";
		$sMultiplicityTip = "";
		
		if (!$this->isCollapsed() && !$this->isCollapsedChild())
		{		
			$sMultiplicityTip = $this->getMultiplicity()->getMultiplicityTip($this);
			if ($this->getMultiplicity() instanceof AnwContentMultiplicity_multiple)
			{
				$sFieldTip = ""; //already written by renderEditInputN!
			}
			else
			{
				$sFieldTip = $this->getFieldTip();
			}
		}
		$sFieldExplain = $this->getFieldExplain();
		
		if ($this instanceof AnwStructuredContentField_tab)
		{
			$sTabDivId = 'tab_'.$this->getInputName($sSuffix);
			return $this->getComponent()->tpl()->renderTabField($sTabDivId, $this->getFieldLabel(), $sInputHtml, $sFieldExplain, $sFieldError);
		}
		else if ($this instanceof AnwStructuredContentField_composed)
		{
			return $this->getComponent()->tpl()->renderComposedField($this->getFieldLabel(), $sInputHtml, $sFieldTip, $sFieldExplain, $sMultiplicityTip, $sFieldError);
		}
		else
		{
			return $this->getComponent()->tpl()->renderInputField($this->getFieldLabel(), $sInputHtml, $sFieldTip, $sFieldExplain, $sMultiplicityTip, $sFieldError);
		}
	}
	
	function renderAdditionalEditInput($sSuffix)
	{
		if (!$this->isMultiple())
		{
			throw new AnwUnexpectedException("JS_AddMultipleContentField on non multiple field");
		}
		
		//temporary disable collapsing
		self::setDoCollapsing(false);
		
		$sDefaultValue = $this->getDefaultValue();
		$nNewSuffixNumber = AnwUtils::genUniqueIdNumeric();
		if ($this instanceof AnwStructuredContentField_composed)
		{
			$sNewSuffix = $this->updateSuffix($sSuffix, $nNewSuffixNumber);
			$oSubContent = $this->rebuildContentFromXml($sDefaultValue);
			$sSubRender = $oSubContent->renderEditHtmlForm(false, "", $sNewSuffix);
			$sDefaultValue = array(AnwStructuredContent::IDX_SUBCONTENT=>$oSubContent, AnwStructuredContent::IDX_RENDERED=>$sSubRender);
		}
		$sHtmlRender = $this->renderEditInputN($sSuffix, $sDefaultValue, $nNewSuffixNumber);
		
		//enable collapsing again
		self::setDoCollapsing(true);
		
		return $sHtmlRender;
	}
	
	/**
	 * Render one instance of a multiple contentfield
	 * //!used by action_edit
	 */
	function renderEditInputN($sSuffix, $sValue, $nInputNumber)
	{
		$sInstancesClass = "instance_".$this->getInputName($sSuffix);
				
		$sRenderedInput = $this->getRenderedEditInput($sValue, $sSuffix, $nInputNumber);
		
		if ($this instanceof AnwStructuredContentField_composed)
		{
			if ($this->getMultiplicity() instanceof AnwContentMultiplicity_multiple)
			{
				$sInstancesName = $this->getInputName($sSuffix);
				$sRenderedInput .= <<<EOF
			
<input type="hidden" name="{$sInstancesName}[]" value="$nInputNumber"/>
EOF;
			}
			$sSuffix = $this->updateSuffix($sSuffix, $nInputNumber); //update suffix
		}
		
		if ($this->isCollapsed() || $this->isCollapsedChild())
		{
			//TODO
			if ($this instanceof AnwStructuredContentField_atomic)
			{
				$HTML = '<span style="padding:0.5em;">'.$sRenderedInput.'</span>';
			}
			else
			{
				//we want to sort contentfields even when they are collapsed!
				//$HTML = $sRenderedInput;
				$HTML = $this->getComponent()->tpl()->renderMultipleInputInstance($sInstancesClass, $this->getMultiplicity()->isSortable(), $sRenderedInput, $this->getFieldLabelSingle(), $this->getFieldTip());
			}
		}
		else
		{
			$HTML = $this->getComponent()->tpl()->renderMultipleInputInstance($sInstancesClass, $this->getMultiplicity()->isSortable(), $sRenderedInput, $this->getFieldLabelSingle(), $this->getFieldTip());
		}
		return $HTML;
	}
	
	protected function getFieldTranslation($sTranslationName, $asParameters=array(), $sTranslationIfEmpty=false)
	{
		$sCacheKey = $sTranslationName.md5(implode(".",$asParameters));
		if (!isset($this->amCachedFieldTranslations[$sCacheKey]))
		{
			//first, try to get it from component
			$sTranslation = $this->doGetFieldTranslation("contentfield", $sTranslationName, $asParameters, true);
			
			//if not found, try to get it from global
			if ($sTranslation === "")
			{
				$sTranslation = $this->doGetFieldTranslation("contentfield", $sTranslationName, $asParameters, false);
			}
			
			//if no translation found
			if ($sTranslation === "")
			{
				if ($sTranslationIfEmpty===false)
				{
					//return #TRANSLATIONID#
					$sTranslation = "#".$this->baseTranslationName("contentfield", 99, true)."#";
				}
				else
				{
					$sTranslation = $sTranslationIfEmpty;
				}
			}
			$this->amCachedFieldTranslations[$sCacheKey] = $sTranslation;
		}
		return $this->amCachedFieldTranslations[$sCacheKey];
	}
	
	/**
	 * Must be overriden by contentfield_page and contentfield_setting.
	 */
	protected abstract function t_structuredcontentfield($id, $asParams=array(), $sLangIfLocal=false, $sValueIfNotFound=false);
	
	/**
	 * Special override system for translations.
	 * Below there is an example applied to contentclass_menu.
	 * C means that the string comes from the class (ex: "menuItem" comes from "AnwContentFieldPage_menuItem")
	 * N means that the string comes from the name of the field (ex: "subitems" is the name of a field in "AnwContentFieldPage_menuItem")
	 * Here is the order for looking to translations.
	 * 
	 * contentfield_menu_items_subitems_link_title_label
	 *               C     N      N      N    N
	 * 
	 * contentfield_menuItem_subitems_link_title_label
	 *                  C       N       N    N
	 * 
	 * contentfield_menuSubItem_link_title_label
	 *                  C        N      N
	 * 
	 * contentfield_link_title_label
	 *                C    N
	 */
	protected final function doGetFieldTranslation($sBasePrefix, $sTranslationName, $asParameters, $bFromComponent, $nMaxDeep=-1)
	{
		if ($nMaxDeep == -1)
		{
			//initialize maxDeep to highest possible value
			$nHighestDeep = 0;
			$oTest = $this;
			while ($oTest->getContainer() instanceof AnwStructuredContentField_composed)
			{
				$oTest = $oTest->getContainer();
				$nHighestDeep++;
			}
			$nMaxDeep = $nHighestDeep;
			
			//try last level translation (special case)
			$sFullTranslationName = $this->baseTranslationName($sBasePrefix, $nMaxDeep, true).$sTranslationName;
			//print "<br/>**Max=".$nMaxDeep.": ".$sFullTranslationName."***<br/>";
			
			if ($bFromComponent)
			{
				$sTranslation = $this->t_structuredcontentfield($sFullTranslationName, $asParameters, false, "");
			}
			else
			{
				$sTranslation = AnwComponent::g_editcontent($sFullTranslationName, $asParameters, false, ""); //instance call required as it's abstract
			}
			if ($sTranslation !== "")
			{
				return $sTranslation;
			}
		}
		
		$sFullTranslationName = $this->baseTranslationName($sBasePrefix, $nMaxDeep, false).$sTranslationName;
		if ($bFromComponent)
		{
			$sTranslation = $this->t_structuredcontentfield($sFullTranslationName, $asParameters, false, "");
		}
		else
		{
			$sTranslation = AnwComponent::g_editcontent($sFullTranslationName, $asParameters, false, ""); //instance call required as it's abstract
		}
		
		//print "<br/>**Max=".$nMaxDeep.": ".$sFullTranslationName."***<br/>";
		if ($sTranslation === "")
		{
			$nMaxDeep--;
			if ($nMaxDeep >= 0)
			{
				$sTranslation = $this->doGetFieldTranslation($sBasePrefix, $sTranslationName, $asParameters, $bFromComponent, $nMaxDeep);
			}
		}
		
		return $sTranslation;
	}
	
	private final function baseTranslationName($sBasePrefix, $nMaxDeep, $bSpecialIncludingContentClass=false)
	{
		//we want something like: contentfield_<classname>_<subfieldname>
		//in example: contentfield_mysqlconnexion_user
		
		if ($nMaxDeep>0 && $this->getContainer() instanceof AnwStructuredContentField_composed) 
		{
			$nNewMaxDeep = $nMaxDeep-1;
			$sTranslationName = $this->getContainer()->baseTranslationName($sBasePrefix, $nNewMaxDeep, $bSpecialIncludingContentClass);
		}
		else if ($bSpecialIncludingContentClass && $this->getContainer() instanceof AnwStructuredContentClass)
		{
			$sClassName = str_replace(array('AnwContentClassPageDefault_','AnwContentClassPageOverride_', 'AnwContentClassSettings'),'',get_class($this->getContainer()));
			$sTranslationName = $sBasePrefix;
			if ($sClassName!="") $sTranslationName .= "_".$sClassName;
		}
		else 
		{
			$sTranslationName = $sBasePrefix;
		}
		
		if ($nMaxDeep==0 && !$bSpecialIncludingContentClass)
		{
			$sTranslationName .= "_".str_replace(array('AnwContentFieldPage_','AnwContentFieldSettings_'),'',get_class($this));
		}
		else
		{
			$sTranslationName .= "_".$this->getName();
		}
		return $sTranslationName;
	}
	
	function getFieldTip($asParameters=array())
	{
		$sTranslation = $this->getFieldTranslation("_tip", $asParameters, "");
		return $sTranslation;
	}
	
	function getFieldExplain($asParameters=array())
	{
		$sTranslation = $this->getFieldTranslation("_explain", $asParameters, "");
		return $sTranslation;
	}
	
	function getFieldLabel()
	{
		$sTranslation = $this->getFieldTranslation("_label");
		return $sTranslation;
	}
	
	function getFieldLabelSingle() 
	{
		$sTranslation = $this->getFieldTranslation("_single");
		return $sTranslation;
	}
	
	function getFieldLabelPlural() 
	{
		$sTranslation = $this->getFieldTranslation("_plural");
		return $sTranslation;
	}
	
	final function isMultiple()
	{
		return ($this->getMultiplicity() instanceof AnwContentMultiplicity_multiple);
	}
	
	
	//TODO : architecture problem, the contentfield shouldnt inherit from AnwStructuredContentFieldsContainer
	//so we just define required abstract functions, but it will never be used
	function getComponentName()
	{
		 throw new AnwUnexpectedException("ContentField is not a component");
	}
		
	static function getComponentsRootDir()
	{
		throw new AnwUnexpectedException("ContentField is not a component");
	}
	static function getComponentsDirsBegin()
	{
		throw new AnwUnexpectedException("ContentField is not a component");
	}
	//due to "stricts oo standards", have to duplicate this function in all components class...
	static function getComponentDir($sName)
	{
		return self::getComponentsRootDir().self::getComponentsDirsBegin().$sName;
	}	
	
	static function getMyComponentType()
	{
		throw new AnwUnexpectedException("ContentField is not a component");
	}
	
	static function discoverEnabledComponents()
	{
		throw new AnwUnexpectedException("ContentField is not a component");
	}
	
	static function loadComponent($sName)
	{
		throw new AnwUnexpectedException("ContentField is not a component");
	}
	
	protected static function debug($sMsg)
	{
		AnwDebug::log("(ContentField)".$sMsg);
	} 
}

/**
 * Indicates that it is a content class (ContentClassPage or ContentClassSettings).
 * We need this for recognize these classes due to architecture problems.
 */
interface AnwStructuredContentClass
{
	
}

/**
 * Indicates that the contentfield isn't composed by any other contentfield (it has its own value).
 */

//warning, this class is used by class_utils
interface AnwStructuredContentField_atomic
{
	function getDefaultValue();
	function setDefaultValue($sValue);
	function hasDefaultValue();
}

/**
 * Indicates that the contentfield is composed by other contentfields.
 */
interface AnwStructuredContentField_composed
{
	
}

/**
 * Indicates that the contentfield is an edition tab.
 */
interface AnwStructuredContentField_tab extends AnwStructuredContentField_composed
{
	
}

//Note : these two interfaces are EXCLUSIVE: a contentfield NEEDS to implements ONE of these two interfaces.  

/**
 * Indicates that the contentfield is always multiple, it can't be single.
 */
interface AnwStructuredContentField_renderedAsMultiple
{
	
}
?>